"""Main module for note-python."""

##
# @mainpage Python Library for the Notecard
#
# @section intro_sec Introduction
# This module contains the core functionality for running the
# note-python library, including the main Notecard class, and
# Serial and I2C sub-classes.
#
# @section dependencies Dependencies
#
# This library requires a physical connection to a Notecard over I2C or
# Serial to be functional.
# @section author Author
#
# Written by Ray Ozzie and Brandon Satrom for Blues Inc.
# @section license License
#
# Copyright (c) 2019 Blues Inc. MIT License. Use of this source code is
# governed by licenses granted by the copyright holder including that found in
# the
# <a href="https://github.com/blues/note-python/blob/master/LICENSE">
#   LICENSE
# </a>
# file.

##
# @file notecard.py
#
# @brief Main module for note-python. Contains core library functionality.

import sys
import json
import time

use_periphery = False
use_micropython = False
use_serial_lock = False
if sys.implementation.name == 'cpython':
    if sys.platform == "linux" or sys.platform == "linux2":
        use_periphery = True
        from periphery import I2C

        use_serial_lock = True
        from filelock import Timeout, FileLock
elif sys.implementation.name == 'micropython':
    use_micropython = True

NOTECARD_I2C_ADDRESS = 0x17

# The notecard is a real-time device that has a fixed size interrupt buffer.
# We can push data at it far, far faster than it can process it,
# therefore we push it in segments with a pause between each segment.
CARD_REQUEST_SEGMENT_MAX_LEN = 250
CARD_REQUEST_SEGMENT_DELAY_MS = 250


def prepareRequest(req, debug=False):
    """Format the request string as a JSON object and add a newline."""
    req_json = json.dumps(req)
    if debug:
        print(req_json)

    req_json += "\n"
    return req_json


def serialReadByte(port):
    """Read a single byte from a Notecard."""
    if sys.implementation.name == 'micropython':
        if not port.any():
            return None
    elif sys.implementation.name == 'cpython':
        if port.in_waiting == 0:
            return None
    return port.read(1)


def serialReset(port):
    """Send a reset command to a Notecard."""
    for i in range(10):
        try:
            port.write(b'\n')
        except:
            continue
        time.sleep(0.5)
        somethingFound = False
        nonControlCharFound = False
        while True:
            data = serialReadByte(port)
            if (data is None) or (data == b''):
                break
            somethingFound = True
            if data[0] >= 0x20:
                nonControlCharFound = True
        if somethingFound and not nonControlCharFound:
            break
        else:
            raise Exception("Notecard not responding")


def serialTransaction(port, req, debug):
    """Perform a single write to and read from a Notecard."""
    req_json = prepareRequest(req, debug)

    seg_off = 0
    seg_left = len(req_json)
    while True:
        seg_len = seg_left
        if seg_len > CARD_REQUEST_SEGMENT_MAX_LEN:
            seg_len = CARD_REQUEST_SEGMENT_MAX_LEN

        port.write(req_json[seg_off:seg_off + seg_len]
                   .encode('utf-8'))
        seg_off += seg_len
        seg_left -= seg_len
        if seg_left == 0:
            break
        time.sleep(CARD_REQUEST_SEGMENT_DELAY_MS / 1000)

    rsp_json = port.readline()

    if debug:
        print(rsp_json.rstrip())

    rsp = json.loads(rsp_json)
    return rsp


def serialCommand(port, req, debug):
    """Perform a single write to and read from a Notecard."""
    req_json = prepareRequest(req, debug)

    seg_off = 0
    seg_left = len(req_json)
    while True:
        seg_len = seg_left
        if seg_len > CARD_REQUEST_SEGMENT_MAX_LEN:
            seg_len = CARD_REQUEST_SEGMENT_MAX_LEN

        port.write(req_json[seg_off:seg_off + seg_len]
                   .encode('utf-8'))
        seg_off += seg_len
        seg_left -= seg_len
        if seg_left == 0:
            break
        time.sleep(CARD_REQUEST_SEGMENT_DELAY_MS / 1000)


class Notecard:
    """Base Notecard class.

    Primary Notecard Class, which provides a shared __init__
    to reset the Notecard via Serial or I2C.
    """

    def __init__(self):
        """Initialize the Notecard through a reset."""
        self.Reset()


class OpenSerial(Notecard):
    """Notecard class for Serial communication."""

    def Command(self, req):
        """Perform a Notecard command and exit with no response."""
        if 'cmd' not in req:
            raise Exception("Please use 'cmd' instead of 'req'")

        if use_serial_lock:
            try:
                self.lock.acquire(timeout=5)
                serialCommand(self.uart, req, self._debug)
            except Timeout:
                raise Exception("Notecard in use")
            finally:
                self.lock.release()
        else:
            serialCommand(self.uart, req, self._debug)

    def Transaction(self, req):
        """Perform a Notecard transaction and return the result."""
        if use_serial_lock:
            try:
                self.lock.acquire(timeout=5)
                return serialTransaction(self.uart, req, self._debug)
            except Timeout:
                raise Exception("Notecard in use")
            finally:
                self.lock.release()
        else:
            return serialTransaction(self.uart, req, self._debug)

    def Reset(self):
        """Reset the Notecard."""
        if use_serial_lock:
            try:
                self.lock.acquire(timeout=5)
                serialReset(self.uart)
            except Timeout:
                raise Exception("Notecard in use")
            finally:
                self.lock.release()
        else:
            serialReset(self.uart)

    def __init__(self, uart_id, debug=False):
        """Initialize the Notecard before a reset."""
        self.uart = uart_id
        self._debug = debug

        if use_serial_lock:
            self.lock = FileLock('serial.lock', timeout=1)
        super().__init__()


class OpenI2C(Notecard):
    """Notecard class for I2C communication."""

    def _sendPayload(self, json):
        chunk_offset = 0
        json_left = len(json)
        sent_in_seg = 0
        while json_left > 0:
            time.sleep(.001)
            chunk_len = min(json_left, self.max)
            reg = bytearray(1)
            reg[0] = chunk_len
            write_data = bytes(json[
                               chunk_offset:
                               chunk_offset + chunk_len
                               ], 'utf-8')
            if use_periphery:
                msgs = [I2C.Message(reg + write_data)]
                self.i2c.transfer(self.addr, msgs)
            else:
                self.i2c.writeto(self.addr, reg + write_data)
            chunk_offset += chunk_len
            json_left -= chunk_len
            sent_in_seg += chunk_len
            if sent_in_seg > CARD_REQUEST_SEGMENT_MAX_LEN:
                sent_in_seg -= CARD_REQUEST_SEGMENT_MAX_LEN
            time.sleep(CARD_REQUEST_SEGMENT_DELAY_MS / 1000)

    def Command(self, req):
        """Perform a Notecard command and exit with no response."""
        if 'cmd' not in req:
            raise Exception("Please use 'cmd' instead of 'req'")

        req_json = prepareRequest(req, self._debug)

        while not self.lock():
            pass

        try:
            self._sendPayload(req_json)
        finally:
            self.unlock()

    def Transaction(self, req):
        """Perform a Notecard transaction and return the result."""
        req_json = prepareRequest(req, self._debug)
        rsp_json = ""

        while not self.lock():
            pass

        try:
            self._sendPayload(req_json)

            chunk_len = 0
            received_newline = False
            start = time.time()
            transaction_timeout_secs = 10
            while True:
                time.sleep(.001)
                reg = bytearray(2)
                reg[0] = 0
                reg[1] = chunk_len
                readlen = chunk_len + 2
                buf = bytearray(readlen)
                if use_periphery:
                    msgs = [I2C.Message(reg), I2C.Message(buf, read=True)]
                    self.i2c.transfer(self.addr, msgs)
                    buf = msgs[1].data
                elif use_micropython:
                    self.i2c.writeto(self.addr, reg, False)
                    self.i2c.readfrom_into(self.addr, buf)
                else:
                    self.i2c.writeto_then_readfrom(self.addr, reg, buf)
                available = buf[0]
                good = buf[1]
                data = buf[2:2 + good]
                if good > 0 and buf[-1] == 0x0a:
                    received_newline = True
                try:
                    rsp_json += "".join(map(chr, data))
                except:
                    pass
                chunk_len = min(available, self.max)
                if chunk_len > 0:
                    continue
                if received_newline:
                    break
                if (time.time() >= start + transaction_timeout_secs):
                    raise Exception("notecard request or response was lost")
                time.sleep(0.05)

        finally:
            self.unlock()

        if self._debug:
            print(rsp_json.rstrip())
        rsp = json.loads(rsp_json)
        return rsp

    def Reset(self):
        """Reset the Notecard."""
        chunk_len = 0

        while not self.lock():
            pass

        try:
            while True:
                time.sleep(.001)
                reg = bytearray(2)
                reg[0] = 0
                reg[1] = chunk_len
                readlen = chunk_len + 2
                buf = bytearray(readlen)
                if use_periphery:
                    msgs = [I2C.Message(reg), I2C.Message(buf, read=True)]
                    self.i2c.transfer(self.addr, msgs)
                    buf = msgs[1].data
                elif use_micropython:
                    self.i2c.writeto(self.addr, reg, False)
                    self.i2c.readfrom_into(self.addr, buf)
                else:
                    self.i2c.writeto_then_readfrom(self.addr, reg, buf)
                available = buf[0]
                if available == 0:
                    break
                chunk_len = min(available, self.max)
        finally:
            self.unlock()

        pass

    def lock(self):
        """Lock the I2C port so the host can interact with the Notecard."""
        if not use_periphery and not use_micropython:
            return self.i2c.try_lock()
        return True

    def unlock(self):
        """Unlock the I2C port."""
        if not use_periphery and not use_micropython:
            return self.i2c.unlock()
        return True

    def __init__(self, i2c, address, max_transfer, debug=False):
        """Initialize the Notecard before a reset."""
        self.i2c = i2c
        self._debug = debug
        if address == 0:
            self.addr = NOTECARD_I2C_ADDRESS
        else:
            self.addr = address
        if max_transfer == 0:
            self.max = 255
        else:
            self.max = max_transfer
        super().__init__()



"""Main Validation decorators for note-python."""

##
# @file validators.py
#
# @brief Validation decorators for note-python.
import functools
import notecard


def validate_card_object(func):
    """Ensure that the passed-in card is a Notecard."""
    @functools.wraps(func)
    def wrap_validator(*args, **kwargs):
        """Check the instance of the passed-in card."""
        card = args[0]
        if not isinstance(card, notecard.Notecard):
            raise Exception("Notecard object required")

        return func(*args, **kwargs)

    return wrap_validator



"""card Fluent API Helper."""

##
# @file card.py
#
# @brief card Fluent API Helper.
#
# @section description Description
# This module contains helper methods for calling card.* Notecard API commands.
# This module is optional and not required for use with the Notecard.

import notecard
import sys
if sys.implementation.name == 'micropython':
    from validators import validate_card_object
else:
    from .validators import validate_card_object


@validate_card_object
def attn(card, mode=None, files=None, seconds=None, payload=None, start=None):
    """Configure interrupt detection between a host and Notecard.

    Args:
        card (Notecard): The current Notecard object.
        mode (string): The attn mode to set.
        files (array): A collection of notefiles to watch.
        seconds (int): A timeout to use when arming attn mode.
        payload (int): When using sleep mode, a payload of data from the host
            that the Notecard should hold in memory until retrieved by
            the host.
        start (bool): When using sleep mode and the host has reawakened,
            request the Notecard to return the stored payload.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "card.attn"}
    if mode:
        req["mode"] = mode
    if files:
        req["files"] = files
    if seconds:
        req["seconds"] = seconds
    if payload:
        req["payload"] = payload
    if start:
        req["start"] = start
    return card.Transaction(req)


@validate_card_object
def time(card):
    """Retrieve the current time and date from the Notecard.

    Args:
        card (Notecard): The current Notecard object.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "card.time"}
    return card.Transaction(req)


@validate_card_object
def status(card):
    """Retrieve the status of the Notecard.

    Args:
        card (Notecard): The current Notecard object.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "card.status"}
    return card.Transaction(req)


@validate_card_object
def temp(card, minutes=None):
    """Retrieve the current temperature from the Notecard.

    Args:
        card (Notecard): The current Notecard object.
        minutes (int): If specified, creates a templated _temp.qo file that
            gathers Notecard temperature value at the specified interval.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "card.temp"}
    if minutes:
        req["minutes"] = minutes
    return card.Transaction(req)


@validate_card_object
def version(card):
    """Retrieve firmware version information from the Notecard.

    Args:
        card (Notecard): The current Notecard object.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "card.version"}
    return card.Transaction(req)


@validate_card_object
def voltage(card, hours=None, offset=None, vmax=None, vmin=None):
    """Retrieve current and historical voltage info from the Notecard.

    Args:
        card (Notecard): The current Notecard object.
        hours (int): Number of hours to analyze.
        offset (int): Number of hours to offset.
        vmax (decimal): max voltage level to report.
        vmin (decimal): min voltage level to report.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "card.voltage"}
    if hours:
        req["hours"] = hours
    if offset:
        req["offset"] = offset
    if vmax:
        req["vmax"] = vmax
    if vmin:
        req["vmin"] = vmin
    return card.Transaction(req)


@validate_card_object
def wireless(card, mode=None, apn=None):
    """Retrieve wireless modem info or customize modem behavior.

    Args:
        card (Notecard): The current Notecard object.
        mode (string): The wireless module mode to set.
        apn (string): Access Point Name (APN) when using an external SIM.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "card.wireless"}
    if mode:
        req["mode"] = mode
    if apn:
        req["apn"] = apn

    return card.Transaction(req)



"""env Fluent API Helper."""

##
# @file env.py
#
# @brief env Fluent API Helper.
#
# @section description Description
# This module contains helper methods for calling env.* Notecard API commands.
# This module is optional and not required for use with the Notecard.

import notecard
import sys
if sys.implementation.name == 'micropython':
    from validators import validate_card_object
else:
    from .validators import validate_card_object


@validate_card_object
def default(card, name=None, text=None):
    """Perform an env.default request against a Notecard.

    Args:
        card (Notecard): The current Notecard object.
        name (string): The name of an environment var to set a default for.
        text (optional): The default value. Omit to delete the default.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "env.default"}
    if name:
        req["name"] = name
    if text:
        req["text"] = text
    return card.Transaction(req)


@validate_card_object
def get(card, name=None):
    """Perform an env.get request against a Notecard.

    Args:
        card (Notecard): The current Notecard object.
        name (string): The name of an environment variable to get.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "env.get"}
    if name:
        req["name"] = name
    return card.Transaction(req)


@validate_card_object
def modified(card):
    """Perform an env.modified request against a Notecard.

    Args:
        card (Notecard): The current Notecard object.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "env.modified"}
    return card.Transaction(req)


@validate_card_object
def set(card, name=None, text=None):
    """Perform an env.set request against a Notecard.

    Args:
        card (Notecard): The current Notecard object.
        name (string): The name of an environment variable to set.
        text (optional): The variable value. Omit to delete.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "env.set"}
    if name:
        req["name"] = name
    if text:
        req["text"] = text
    return card.Transaction(req)



"""note Fluent API Helper."""

##
# @file note.py
#
# @brief note Fluent API Helper.
#
# @section description Description
# This module contains helper methods for calling note.* Notecard API commands.
# This module is optional and not required for use with the Notecard.

import notecard
import sys
if sys.implementation.name == 'micropython':
    from validators import validate_card_object
else:
    from .validators import validate_card_object


@validate_card_object
def add(card, file=None, body=None, payload=None, sync=None):
    """Add a Note to a Notefile.

    Args:
        card (Notecard): The current Notecard object.
        file (string): The name of the file.
        body (JSON object): A developer-defined tracker ID.
        payload (string): An optional base64-encoded string.
        sync (bool): Perform an immediate sync after adding.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "note.add"}
    if file:
        req["file"] = file
    if body:
        req["body"] = body
    if payload:
        req["payload"] = payload
    if sync is not None:
        req["sync"] = sync
    return card.Transaction(req)


@validate_card_object
def changes(card, file=None, tracker=None, maximum=None,
            start=None, stop=None, deleted=None, delete=None):
    """Incrementally retrieve changes within a Notefile.

    Args:
        card (Notecard): The current Notecard object.
        file (string): The name of the file.
        tracker (string): A developer-defined tracker ID.
        maximum (int): Maximum number of notes to return.
        start (bool): Should tracker be reset to the beginning
            before a get.
        stop (bool): Should tracker be deleted after get.
        deleted (bool): Should deleted notes be returned.
        delete (bool): Should notes in a response be auto-deleted.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "note.changes"}
    if file:
        req["file"] = file
    if tracker:
        req["tracker"] = tracker
    if maximum:
        req["max"] = maximum
    if start is not None:
        req["start"] = start
    if stop is not None:
        req["stop"] = stop
    if deleted is not None:
        req["deleted"] = deleted
    if delete is not None:
        req["delete"] = delete
    return card.Transaction(req)


@validate_card_object
def get(card, file="data.qi", note_id=None, delete=None, deleted=None):
    """Retrieve a note from an inbound or DB Notefile.

    Args:
        card (Notecard): The current Notecard object.
        file (string): The inbound or DB notefile to retrieve a
            Notefile from.
        note_id (string): (DB files only) The ID of the note to retrieve.
        delete (bool): Whether to delete the note after retrieval.
        deleted (bool): Whether to allow retrieval of a deleted note.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "note.get"}
    req["file"] = file
    if note_id:
        req["note"] = note_id
    if delete is not None:
        req["delete"] = delete
    if deleted is not None:
        req["deleted"] = deleted
    return card.Transaction(req)


@validate_card_object
def delete(card, file=None, note_id=None):
    """Delete a DB note in a Notefile by its ID.

    Args:
        card (Notecard): The current Notecard object.
        file (string): The file name of the DB notefile.
        note_id (string): The id of the note to delete.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "note.delete"}
    if file:
        req["file"] = file
    if note_id:
        req["note"] = note_id
    return card.Transaction(req)


@validate_card_object
def update(card, file=None, note_id=None, body=None, payload=None):
    """Update a note in a DB Notefile by ID.

    Args:
        card (Notecard): The current Notecard object.
        file (string): The file name of the DB notefile.
        note_id (string): The id of the note to update.
        body (JSON): The JSON object to add to the note.
        payload (string): The base64-encoded JSON payload to
            add to the note.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "note.get"}
    if file:
        req["file"] = file
    if note_id:
        req["note"] = note_id
    if body:
        req["body"] = body
    if payload:
        req["payload"] = payload
    return card.Transaction(req)


@validate_card_object
def template(card, file=None, body=None, length=None):
    """Create a template for new Notes in a Notefile.

    Args:
        card (Notecard): The current Notecard object.
        file (string): The file name of the notefile.
        body (JSON): A sample JSON body that specifies field names and
            values as "hints" for the data type.
        length (int): If provided, the maximum length of a payload that
            can be sent in Notes for the template Notefile.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "note.get"}
    if file:
        req["file"] = file
    if body:
        req["body"] = body
    if length:
        req["length"] = length
    return card.Transaction(req)



import os
import sys
import serial
import periphery
import pytest

from unittest.mock import Mock, MagicMock, patch

sys.path.insert(0, os.path.abspath(
                os.path.join(os.path.dirname(__file__), '..')))

import notecard  # noqa: E402
from notecard import card, hub, note, env, file  # noqa: E402


def get_serial_and_port():
    serial = Mock()  # noqa: F811
    port = serial.Serial("/dev/tty.foo", 9600)
    port.read.side_effect = [b'\r', b'\n', None]
    port.readline.return_value = "\r\n"

    nCard = notecard.OpenSerial(port)

    return (nCard, port)


def get_i2c_and_port():
    periphery = Mock()  # noqa: F811
    port = periphery.I2C("dev/i2c-foo")
    port.try_lock.return_value = True

    nCard = notecard.OpenI2C(port, 0x17, 255)

    return (nCard, port)


def test_open_serial():
    nCard, _ = get_serial_and_port()

    assert nCard.uart is not None


def test_open_i2c():
    nCard, _ = get_i2c_and_port()

    assert nCard.i2c is not None


def test_transaction():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"connected\":true}\r\n"
    print(port.readline())
    response = nCard.Transaction({"req": "hub.status"})

    assert "connected" in response
    assert response["connected"] is True


def test_command():
    nCard, port = get_serial_and_port()

    response = nCard.Command({"cmd": "card.sleep"})

    assert response is None


def test_command_fail_if_req():
    nCard, port = get_serial_and_port()

    with pytest.raises(Exception, match="Please use 'cmd' instead of 'req'"):
        nCard.Command({"req": "card.sleep"})


def test_hub_set():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{}\r\n"
    response = hub.set(nCard, product="com.blues.tester",
                       sn="foo",
                       mode="continuous",
                       outbound=2,
                       inbound=60,
                       duration=5,
                       sync=True,
                       align=True,
                       voutbound="2.3",
                       vinbound="3.3",
                       host="http://hub.blues.foo")

    assert response == {}


def test_hub_set_invalid_card():
    with pytest.raises(Exception, match="Notecard object required"):
        hub.set(None, product="com.blues.tester")


def test_hub_sync():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{}\r\n"
    response = hub.sync(nCard)

    assert response == {}


def test_hub_sync_status():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"status\":\"connected\"}\r\n"

    response = hub.syncStatus(nCard, True)

    assert "status" in response
    assert response["status"] == "connected"


def test_hub_status():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"connected\":true}\r\n"

    response = hub.status(nCard)

    assert "connected" in response
    assert response["connected"] is True


def test_hub_log():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{}\r\n"

    response = hub.log(nCard, "there's been an issue!", False)

    assert response == {}


def test_hub_get():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"mode\":\"continuous\"}\r\n"

    response = hub.get(nCard)

    assert "mode" in response
    assert response["mode"] == "continuous"


def test_card_time():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"time\":1592490375}\r\n"

    response = card.time(nCard)

    assert "time" in response
    assert response["time"] == 1592490375


def test_card_status():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"usb\":true,\"status\":\"{normal}\"}\r\n"

    response = card.status(nCard)

    assert "status" in response
    assert response["status"] == "{normal}"


def test_card_temp():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"value\":33.625,\"calibration\":-3.0}\r\n"

    response = card.temp(nCard, minutes=20)

    assert "value" in response
    assert response["value"] == 33.625


def test_card_attn():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"set\":true}\r\n"

    response = card.attn(nCard, mode="arm, files",
                         files=["sensors.qo"],
                         seconds=10, payload={"foo": "bar"},
                         start=True)

    assert "set" in response
    assert response["set"] is True


def test_card_attn_with_invalid_card():
    with pytest.raises(Exception, match="Notecard object required"):
        card.attn(None, mode="arm")


def test_card_voltage():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"hours\":707}\r\n"

    response = card.voltage(nCard, hours=24, offset=5, vmax=4, vmin=3)

    assert "hours" in response
    assert response["hours"] == 707


def test_card_wireless():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"status\":\"{modem-off}\",\"count\":1}\r\n"

    response = card.wireless(nCard, mode="auto", apn="-")

    assert "status" in response
    assert response["status"] == "{modem-off}"


def test_card_version():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"version\":\"notecard-1.2.3.9950\"}\r\n"

    response = card.version(nCard)

    assert "version" in response
    assert response["version"] == "notecard-1.2.3.9950"


def test_note_add():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"total\":1}\r\n"

    response = note.add(nCard, file="sensors.qo",
                        body={"temp": 72.22},
                        payload="b64==",
                        sync=True)

    assert "total" in response
    assert response["total"] == 1


def test_note_get():
    nCard, port = get_serial_and_port()

    port.readline.return_value = \
        "{\"note\":\"s\",\"body\":{\"s\":\"foo\"}}\r\n"

    response = note.get(nCard, file="settings.db",
                        note_id="s",
                        delete=True,
                        deleted=False)

    assert "note" in response
    assert response["note"] == "s"


def test_note_delete():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{}\r\n"

    response = note.delete(nCard, file="settings.db", note_id="s")

    assert response == {}


def test_note_update():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{}\r\n"

    response = note.update(nCard, file="settings.db", note_id="s",
                           body={"foo": "bar"}, payload="123dfb==")

    assert response == {}


def test_note_changes():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"changes\":5,\"total\":15}\r\n"

    response = note.changes(nCard, file="sensors.qo",
                            tracker="123",
                            maximum=10,
                            start=True,
                            stop=False,
                            deleted=False,
                            delete=True)

    assert "changes" in response
    assert response["changes"] == 5


def test_note_template():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"bytes\":40}\r\n"

    response = note.template(nCard, file="sensors.qo",
                             body={"temp": 1.1, "hu": 1},
                             length=5)

    assert "bytes" in response
    assert response["bytes"] == 40


def test_debug_mode_on_serial():
    serial = Mock()  # noqa: F811
    port = serial.Serial("/dev/tty.foo", 9600)
    port.read.side_effect = [b'\r', b'\n', None]

    nCard = notecard.OpenSerial(port, debug=True)

    assert nCard._debug


def test_debug_mode_on_i2c():
    periphery = Mock()  # noqa: F811
    port = periphery.I2C("dev/i2c-foo")
    port.try_lock.return_value = True

    nCard = notecard.OpenI2C(port, 0x17, 255, debug=True)

    assert nCard._debug


def test_env_default():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{}\r\n"

    response = env.default(nCard, name="pump", text="on")

    assert response == {}


def test_env_set():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{}\r\n"

    response = env.set(nCard, name="pump", text="on")

    assert response == {}


def test_env_get():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{}\r\n"

    response = env.get(nCard, name="pump")

    assert response == {}


def test_env_modified():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"time\": 1605814493}\r\n"

    response = env.modified(nCard)

    assert "time" in response
    assert response["time"] == 1605814493


def test_file_delete():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{}\r\n"

    response = file.delete(nCard, files=["sensors.qo"])

    assert response == {}


def test_file_changes():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"total\":5}\r\n"

    response = file.changes(nCard, tracker="123", files=["sensors.qo"])

    assert "total" in response
    assert response["total"] == 5


def test_file_stats():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"total\":24}\r\n"

    response = file.stats(nCard)

    assert "total" in response
    assert response["total"] == 24


def test_file_pendingChanges():
    nCard, port = get_serial_and_port()

    port.readline.return_value = "{\"changes\":1}\r\n"

    response = file.pendingChanges(nCard)

    assert "changes" in response
    assert response["changes"] == 1



"""hub Fluent API Helper."""

##
# @file hub.py
#
# @brief hub Fluent API Helper.
#
# @section description Description
# This module contains helper methods for calling hub.* Notecard API commands.
# This module is optional and not required for use with the Notecard.

import notecard
import sys
if sys.implementation.name == 'micropython':
    from validators import validate_card_object
else:
    from .validators import validate_card_object


@validate_card_object
def set(card, product=None, sn=None, mode=None, outbound=None,
        inbound=None, duration=None, sync=False, align=None, voutbound=None,
        vinbound=None, host=None):
    """Configure Notehub behavior on the Notecard.

    Args:
        card (Notecard): The current Notecard object.
        product (string): The ProductUID of the project.
        sn (string): The Serial Number of the device.
        mode (string): The sync mode to use.
        outbound (int): Max time to wait to sync outgoing data.
        inbound (int): Max time to wait to sync incoming data.
        duration (int): If in continuous mode, the amount of time, in minutes,
            of each session.
        sync (bool): If in continuous mode, whether to automatically
            sync each time a change is detected on the device or Notehub.
        align (bool): To align syncs to a regular time-interval, as opposed
            to using max time values.
        voutbound (string): Overrides "outbound" with a voltage-variable value.
        vinbound (string): Overrides "inbound" with a voltage-variable value.
        host (string): URL of an alternative or private Notehub instance.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "hub.set"}
    if product:
        req["product"] = product
    if sn:
        req["sn"] = sn
    if mode:
        req["mode"] = mode
    if outbound:
        req["outbound"] = outbound
    if inbound:
        req["inbound"] = inbound
    if duration:
        req["duration"] = duration
    if sync is not None:
        req["sync"] = sync
    if align is not None:
        req["align"] = align
    if voutbound:
        req["voutbound"] = voutbound
    if vinbound:
        req["vinbound"] = vinbound
    if host:
        req["host"] = host

    return card.Transaction(req)


@validate_card_object
def sync(card):
    """Initiate a sync of the Notecard to Notehub.

    Args:
        card (Notecard): The current Notecard object.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "hub.sync"}
    return card.Transaction(req)


@validate_card_object
def syncStatus(card, sync=None):
    """Retrieve the status of a sync request.

    Args:
        card (Notecard): The current Notecard object.
        sync (bool): True if sync should be auto-initiated pending
            outbound data.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "hub.sync.status"}
    if sync is not None:
        req["sync"] = sync

    return card.Transaction(req)


@validate_card_object
def status(card):
    """Retrieve the status of the Notecard's connection.

    Args:
        card (Notecard): The current Notecard object.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "hub.status"}
    return card.Transaction(req)


@validate_card_object
def log(card, text, alert=False, sync=False):
    """Send a log request to the Notecard.

    Args:
        card (Notecard): The current Notecard object.
        text (string): The ProductUID of the project.
        alert (bool): True if the message is urgent.
        sync (bool): Whether to sync right away.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "hub.log"}
    req["text"] = text
    req["alert"] = alert
    req["sync"] = sync
    return card.Transaction(req)


@validate_card_object
def get(card):
    """Retrieve the current Notehub configuration parameters.

    Args:
        card (Notecard): The current Notecard object.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "hub.get"}
    return card.Transaction(req)



"""file Fluent API Helper."""

##
# @file file.py
#
# @brief file Fluent API Helper.
#
# @section description Description
# This module contains helper methods for calling file.* Notecard API commands.
# This module is optional and not required for use with the Notecard.

import notecard
import sys
if sys.implementation.name == 'micropython':
    from validators import validate_card_object
else:
    from .validators import validate_card_object


@validate_card_object
def changes(card, tracker=None, files=None):
    """Perform individual or batch queries on Notefiles.

    Args:
        card (Notecard): The current Notecard object.
        tracker (string): A developer-defined tracker ID.
        files (array): A list of Notefiles to retrieve changes for.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "file.changes"}
    if tracker:
        req["tracker"] = tracker
    if files:
        req["files"] = files
    return card.Transaction(req)


@validate_card_object
def delete(card, files=None):
    """Delete individual notefiles and their contents.

    Args:
        card (Notecard): The current Notecard object.
        files (array): A list of Notefiles to delete.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "file.delete"}
    if files:
        req["files"] = files
    return card.Transaction(req)


@validate_card_object
def stats(card):
    """Obtain statistics about local notefiles.

    Args:
        card (Notecard): The current Notecard object.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "file.stats"}

    return card.Transaction(req)


@validate_card_object
def pendingChanges(card):
    """Retrieve information about pending Notehub changes.

    Args:
        card (Notecard): The current Notecard object.

    Returns:
        string: The result of the Notecard request.
    """
    req = {"req": "file.changes.pending"}

    return card.Transaction(req)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="note-python",
    version="1.4.0",
    author="Blues Inc.",
    author_email="support@blues.com",
    description="Cross-platform Python Library for the Blues Wireless Notecard,",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/blues/note-python",
    packages=setuptools.find_packages(),
    license="MIT",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Natural Language :: English",
    ],
    install_requires=["filelock"],
    python_requires='>=3.6',
)